# USBGuard Simple GUI Py/Qt
# Copyright (C) 2019  Marco Nicola
#
# This file is part of "USBGuard Simple GUI Py/Qt".
#
# "USBGuard Simple GUI Py/Qt" is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# "USBGuard Simple GUI Py/Qt" is distributed in the hope that it will be
# useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with "USBGuard Simple GUI Py/Qt".  If not, see
# <https://www.gnu.org/licenses/>.

import os
import signal
import sys
from PySide2.QtGui import QIcon
from PySide2.QtWidgets import (QAction,
                               QApplication,
                               QMenu,
                               QMessageBox,
                               QSystemTrayIcon)
from . import APP_NAME
from .device import Device
from .main_window import MainWindow
from .rules import RuleTarget
from .usbguard_dbus_interface import (CallbackEventType,
                                      EventPresenceChangeType,
                                      UsbguardDbusInterface)

SYSTEM_TRAY_APP_NAME = f'{APP_NAME} - System Tray App'


class SystemTrayApp:
    def __init__(
        self,
        app: QApplication,
        usbguard_dbus: UsbguardDbusInterface
    ) -> None:
        self._app = app
        self._main_window = MainWindow(app, usbguard_dbus)
        self._open_action = self._create_open_action()
        self._quit_action = self._create_quit_action()
        self._menu = self._create_menu()
        self._tray_icon = self._create_tray_icon()

        self._usbguard_dbus = usbguard_dbus
        self._register_dbus_callbacks()

    def start(self) -> None:
        if not QSystemTrayIcon.isSystemTrayAvailable():
            QMessageBox.critical(
                None,
                SYSTEM_TRAY_APP_NAME,
                'System tray is not available.')
            sys.exit(1)

        if not QSystemTrayIcon.supportsMessages():
            QMessageBox.warning(
                None,
                SYSTEM_TRAY_APP_NAME,
                'The system tray does not support notification '
                '("balloon") messages.')

        self._tray_icon.show()

    def _create_open_action(self) -> QAction:
        action = QAction('Open')
        action.triggered.connect(self._open_window)
        return action

    def _create_quit_action(self) -> QAction:
        action = QAction('Quit')
        action.triggered.connect(self._quit_app)
        return action

    def _create_menu(self) -> QMenu:
        menu = QMenu()
        menu.addAction(self._open_action)
        menu.addSeparator()
        menu.addAction(self._quit_action)
        return menu

    def _create_tray_icon(self):
        icon_filename = os.path.join(os.path.dirname(__file__), 'icon.svg')
        icon = QIcon(icon_filename)
        tray_icon = QSystemTrayIcon()
        tray_icon.setContextMenu(self._menu)
        tray_icon.setIcon(icon)
        tray_icon.activated.connect(self._on_activated)
        tray_icon.messageClicked.connect(self._open_window)
        tray_icon.setToolTip(APP_NAME)
        return tray_icon

    def _on_activated(self, reason: QSystemTrayIcon.ActivationReason) -> None:
        if reason in (QSystemTrayIcon.DoubleClick, QSystemTrayIcon.Trigger):
            self._open_window()

    def _quit_app(self) -> None:
        self._app.quit()

    def _open_window(self) -> None:
        self._main_window.show()

    def _register_dbus_callbacks(self) -> None:
        if not QSystemTrayIcon.supportsMessages():
            # The user will be notified by `start()`
            return

        self._usbguard_dbus.register_callback(
            CallbackEventType.DEVICE_PRESENCE_CHANGED,
            self._on_device_presence_changed)

        self._usbguard_dbus.register_callback(
            CallbackEventType.DEVICE_POLICY_CHANGED,
            self._on_device_policy_changed)

        self._usbguard_dbus.register_callback(
            CallbackEventType.DEVICE_PRESENCE_CHANGED_ERROR,
            self._on_device_presence_changed_error)

        self._usbguard_dbus.register_callback(
            CallbackEventType.DEVICE_POLICY_CHANGED_ERROR,
            self._on_device_policy_changed_error)

    def _on_device_presence_changed(
        self,
        device: Device,
        event: EventPresenceChangeType,
        _target: int
    ) -> None:
        if event is EventPresenceChangeType.REMOVE:
            self._show_removed_device_message(device)
        elif device.rule.target is RuleTarget.ALLOW:
            self._show_allowed_device_message(device)
        else:
            self._show_device_to_be_managed_message(device)

    def _on_device_policy_changed(
        self,
        device: Device,
        target_old: RuleTarget,
        target_new: RuleTarget,
        rule_id: int,
    ) -> None:
        self._tray_icon.showMessage(
            f'USB device policy changed for "{device.human_readable_name}"',
            f'Rule #{rule_id}: {target_old.value} →  {target_new.value}\n\n'
            f'{device.rule.human_repr}',
            QSystemTrayIcon.Information)

    def _show_removed_device_message(self, device: Device):
        self._tray_icon.showMessage(
            f'USB device "{device.human_readable_name}" was removed',
            device.rule.human_repr,
            QSystemTrayIcon.Information)

    def _show_allowed_device_message(self, device: Device):
        self._tray_icon.showMessage(
            f'USB device "{device.human_readable_name}" was allowed',
            device.rule.human_repr,
            QSystemTrayIcon.Information)

    def _show_device_to_be_managed_message(self, device: Device):
        self._tray_icon.showMessage(
            f'USB device "{device.human_readable_name}" inserted',
            f'Click here to open {APP_NAME} and take action.\n\n'
            f'{device.rule.human_repr}',
            QSystemTrayIcon.Warning,
            100_000_000)

    def _on_device_presence_changed_error(self, error: Exception) -> None:
        self._show_error_message(
            'Error while processing a '
            'device presence change event.',
            error)

    def _on_device_policy_changed_error(self, error: Exception) -> None:
        self._show_error_message(
            'An error occurred while processing a '
            'device policy change event.',
            error)

    def _show_error_message(self, description: str, error: Exception) -> None:
        self._tray_icon.showMessage(
            f'{APP_NAME} - Application Error',
            f'{description}\n\nDetails:\n{type(error).__name__} - {error}',
            QSystemTrayIcon.Critical,
            100_000_000)


def main() -> None:
    signal.signal(signal.SIGINT, signal.SIG_DFL)

    app = QApplication(sys.argv)
    app.setQuitOnLastWindowClosed(False)

    usbguard_dbus = UsbguardDbusInterface()

    system_tray_app = SystemTrayApp(app, usbguard_dbus)
    system_tray_app.start()

    sys.exit(app.exec_())


if __name__ == '__main__':
    main()



# USBGuard Simple GUI Py/Qt
# Copyright (C) 2019  Marco Nicola
#
# This file is part of "USBGuard Simple GUI Py/Qt".
#
# "USBGuard Simple GUI Py/Qt" is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# "USBGuard Simple GUI Py/Qt" is distributed in the hope that it will be
# useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with "USBGuard Simple GUI Py/Qt".  If not, see
# <https://www.gnu.org/licenses/>.

from enum import auto, Enum, IntEnum, unique
from typing import Callable, Dict, List, Set, Optional
from dbus import (Array,
                  Dictionary,
                  Interface,
                  String,
                  Struct,
                  SystemBus,
                  UInt32)
from dbus.mainloop.glib import DBusGMainLoop

from usbguard_simple_gui_py_qt.rules import RuleTarget
from .device import Device
from .rule_parsing import RuleParser

_BUS_NAME = 'org.usbguard1'
_POLICY_PATH = '/org/usbguard1/Policy'
_POLICY_IFACE_NAME = 'org.usbguard.Policy1'
_DEVICES_PATH = '/org/usbguard1/Devices'
_DEVICES_IFACE_NAME = 'org.usbguard.Devices1'


@unique
class CallbackEventType(Enum):
    DEVICE_PRESENCE_CHANGED = auto()
    DEVICE_PRESENCE_CHANGED_ERROR = auto()
    DEVICE_POLICY_CHANGED = auto()
    DEVICE_POLICY_CHANGED_ERROR = auto()


@unique
class EventPresenceChangeType(IntEnum):
    PRESENT = 0
    INSERT = 1
    UPDATE = 2
    REMOVE = 3


_TARGET_TO_INT = {
    RuleTarget.ALLOW: 0,
    RuleTarget.BLOCK: 1,
    RuleTarget.REJECT: 2,
}

_INT_TO_TARGET = {_TARGET_TO_INT[t]: t for t in _TARGET_TO_INT}


class UsbguardDbusInterface:

    def __init__(self) -> None:
        DBusGMainLoop(set_as_default=True)

        self._bus = SystemBus()

        self._policy_proxy = self._bus.get_object(_BUS_NAME, _POLICY_PATH)
        self._policy = Interface(self._policy_proxy, _POLICY_IFACE_NAME)

        self._devices_proxy = self._bus.get_object(_BUS_NAME, _DEVICES_PATH)
        self._devices = Interface(self._devices_proxy, _DEVICES_IFACE_NAME)

        self._devices.connect_to_signal(
            'DevicePresenceChanged', self._on_device_presence_changed)

        self._devices.connect_to_signal(
            'DevicePolicyChanged', self._on_device_policy_changed)

        self._callbacks: Dict[CallbackEventType, Set[Callable]] = \
            {e: set() for e in CallbackEventType}

    def register_callback(
        self,
        event_type: CallbackEventType,
        callback: Callable
    ) -> None:
        callbacks = self._callbacks[event_type]
        if callback in callbacks:
            raise Exception(f'callback {callback} already registered')
        self._callbacks[event_type].add(callback)

    def unregister_callback(
        self,
        event_type: CallbackEventType,
        callback: Callable
    ) -> None:
        self._callbacks[event_type].remove(callback)

    def list_devices(self, query: str = 'match') -> List[Device]:
        response: Array[Struct[UInt32, String]] = \
            self._devices.listDevices(query)
        return [
            Device(
                device_id=int(device_struct[0]),
                rule=RuleParser.parse(str(device_struct[1])))
            for device_struct in response
        ]

    def apply_device_policy(
        self,
        device_id: int,
        target: RuleTarget,
        permanent: bool
    ) -> Optional[int]:
        response: UInt32 = self._devices.applyDevicePolicy(
            device_id,
            _TARGET_TO_INT[target],
            permanent)
        return int(response) if permanent else None

    def _on_device_presence_changed(
        self,
        device_id: UInt32,
        event: UInt32,
        target: UInt32,
        device_rule: String,
        _attributes: Dictionary
    ) -> None:
        """
        :param device_id: Device id of the device.
        :param event: Type of the presence change event in numerical form.
            0 = Present, 1 = Insert, 2 = Update, 3 = Remove.
        :param target: The current authorization target of the device in
            numerical form.
        :param device_rule: Device specific rule.
        :param _attributes: A dictionary of device attributes and their values.
        """
        try:
            resolved_event = EventPresenceChangeType(int(event))
            resolved_target = int(target)
            device = Device(
                device_id=int(device_id),
                rule=RuleParser.parse(device_rule))
        except Exception as error:
            event_type = CallbackEventType.DEVICE_PRESENCE_CHANGED_ERROR
            for callback in self._callbacks[event_type]:
                callback(error)
            return

        event_type = CallbackEventType.DEVICE_PRESENCE_CHANGED
        for callback in self._callbacks[event_type]:
            callback(device, resolved_event, resolved_target)

    def _on_device_policy_changed(
        self,
        device_id: UInt32,
        target_old: UInt32,
        target_new: UInt32,
        device_rule: String,
        rule_id: UInt32,
        _attributes: Dictionary
    ) -> None:
        """
        :param device_id: Device id of the device.
        :param target_old: Previous authorization target in numerical form.
        :param target_new: Current authorization target in numerical form.
        :param device_rule: Device specific rule.
        :param rule_id: A rule id of the matched rule. Otherwise a reserved
            rule id value is used.
        :param _attributes: A dictionary of device attributes and their values.
        """
        try:
            resolved_target_old = _INT_TO_TARGET[int(target_old)]
            resolved_target_new = _INT_TO_TARGET[int(target_new)]
            resolved_rule_id = int(rule_id)
            device = Device(
                device_id=int(device_id),
                rule=RuleParser.parse(device_rule))
        except Exception as error:
            event_type = CallbackEventType.DEVICE_POLICY_CHANGED_ERROR
            for callback in self._callbacks[event_type]:
                callback(error)
            return

        event_type = CallbackEventType.DEVICE_POLICY_CHANGED
        for callback in self._callbacks[event_type]:
            callback(
                device,
                resolved_target_old,
                resolved_target_new,
                resolved_rule_id)



# USBGuard Simple GUI Py/Qt
# Copyright (C) 2019  Marco Nicola
#
# This file is part of "USBGuard Simple GUI Py/Qt".
#
# "USBGuard Simple GUI Py/Qt" is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# "USBGuard Simple GUI Py/Qt" is distributed in the hope that it will be
# useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with "USBGuard Simple GUI Py/Qt".  If not, see
# <https://www.gnu.org/licenses/>.

from dataclasses import dataclass, field
from enum import Enum, unique
from typing import Dict, Generic, List, Optional, TypeVar


@unique
class DeviceAttributeName(Enum):
    ID = 'id'
    HASH = 'hash'
    PARENT_HASH = 'parent-hash'
    NAME = 'name'
    SERIAL = 'serial'
    VIA_PORT = 'via-port'
    WITH_INTERFACE = 'with-interface'
    WITH_CONNECT_TYPE = 'with-connect-type'


@unique
class DeviceAttributeOperator(Enum):
    ALL_OF = 'all-of'
    ONE_OF = 'one-of'
    NONE_OF = 'none-of'
    EQUALS = 'equals'
    EQUALS_ORDERED = 'equals-ordered'


@unique
class RuleTarget(Enum):
    ALLOW = 'allow'
    BLOCK = 'block'
    REJECT = 'reject'


DeviceAttributeValueType = TypeVar('DeviceAttributeValueType')
@dataclass
class DeviceAttribute(Generic[DeviceAttributeValueType]):
    name: DeviceAttributeName
    operator: Optional[DeviceAttributeOperator]
    values: List[DeviceAttributeValueType]

    @property
    def human_repr(self) -> str:
        if len(self.values) == 1 and self.operator is None:
            right_value = self._value_human_repr(self.values[0])
        else:
            right_value = '%s{ %s }' % (
                self.operator.value if self.operator else '',
                ' '.join(map(self._value_human_repr, self.values)))
        return f'{self.name.value}: {right_value}'

    @staticmethod
    def _value_human_repr(value) -> str:
        return f'"{value}"' if isinstance(value, str) else str(value)


@dataclass
class DeviceId:
    vendor_id: Optional[int]
    product_id: Optional[int]

    def __repr__(self) -> str:
        return f'{self.vendor_repr}:{self.product_repr}'

    @property
    def vendor_repr(self) -> str:
        return self._hex_repr(self.vendor_id)

    @property
    def product_repr(self) -> str:
        return self._hex_repr(self.product_id)

    @staticmethod
    def _hex_repr(value: Optional[int]) -> str:
        return '*' if value is None else '%04x' % value


@dataclass
class DeviceInterfaceType:
    iface_class: int
    iface_subclass: Optional[int]
    iface_protocol: Optional[int]

    def __repr__(self) -> str:
        return f'{self.cc}:{self.ss}:{self.pp}'

    @property
    def cc(self) -> str:
        return '%02x' % self.iface_class

    @property
    def ss(self) -> str:
        return self._hex_repr(self.iface_subclass)

    @property
    def pp(self) -> str:
        return self._hex_repr(self.iface_protocol)

    @staticmethod
    def _hex_repr(value: Optional[int]) -> str:
        return '*' if value is None else '%02x' % value


@dataclass
class Rule:
    target: RuleTarget
    attributes: Dict[DeviceAttributeName, DeviceAttribute] = \
        field(default_factory=dict)

    def __getattr__(self, name: str) -> Optional[DeviceAttribute]:
        enum_name = name.replace('_', '-')
        try:
            attribute = DeviceAttributeName(enum_name)
        except ValueError:
            raise AttributeError(
                f"'{type(self).__name__}' object has no attribute '{name}'")
        return self.attributes.get(attribute)

    @property
    def human_repr(self) -> str:
        return '\n'.join([
            f'target: {self.target.value}',
            *[
                # Iterating on the enum, and not on the attributes directly,
                # to ensure that the same order is always respected
                self.attributes[key].human_repr
                for key in DeviceAttributeName
                if key in self.attributes
            ]
        ])



# USBGuard Simple GUI Py/Qt
# Copyright (C) 2019  Marco Nicola
#
# This file is part of "USBGuard Simple GUI Py/Qt".
#
# "USBGuard Simple GUI Py/Qt" is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# "USBGuard Simple GUI Py/Qt" is distributed in the hope that it will be
# useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with "USBGuard Simple GUI Py/Qt".  If not, see
# <https://www.gnu.org/licenses/>.

from string import ascii_lowercase, whitespace
from typing import List, Optional, Union
from .rules import (DeviceAttribute,
                    DeviceAttributeName,
                    DeviceAttributeOperator,
                    DeviceId,
                    DeviceInterfaceType,
                    Rule,
                    RuleTarget)


class RuleParsingError(Exception):
    pass


class RuleParser:
    @staticmethod
    def parse(value: str) -> Rule:
        parser = RuleParser(value)
        parser._parse()
        return parser._rule

    def __init__(self, value: str) -> None:
        self._input_value: str = value
        self._value: str = value
        self._rule: Optional[Rule] = None

    _RESERVED_WORD_CHARS = f'{ascii_lowercase}-'

    def _parse(self) -> None:
        self._consume_optional_whitespaces()
        self._consume_target_and_init_rule()
        self._consume_attributes()

    def _consume_target_and_init_rule(self) -> None:
        word = self._consume_reserved_word()
        try:
            target = RuleTarget(word)
            self._rule = Rule(target=target)
        except ValueError as error:
            self._raise_error(str(error))

    def _consume_attributes(self) -> None:
        self._consume_optional_whitespaces()
        while self._value:
            self._consume_attribute()
            self._consume_optional_whitespaces()

    def _consume_attribute(self):
        name = self._consume_attribute_name()

        if name in self._rule.attributes:
            self._raise_error(f'attribute {name} already set')

        self._consume_mandatory_whitespaces()

        operator = self._consume_optional_operator()

        if operator:
            self._consume_mandatory_whitespaces()
            values = self._consume_multi_attributes_values(name)
        else:
            values = self._consume_single_or_multi_attribute_values(name)

        attribute = DeviceAttribute(name, operator, values)
        self._rule.attributes[name] = attribute

    def _consume_attribute_name(self) -> DeviceAttributeName:
        word = self._consume_reserved_word()

        try:
            return DeviceAttributeName(word)
        except ValueError as error:
            self._raise_error(str(error))

    def _consume_optional_operator(self) -> Optional[DeviceAttributeOperator]:
        for operator in DeviceAttributeOperator:
            word = operator.value
            if self._can_consume_exact_word(word):
                self._yank(len(word))
                return operator
        return None

    def _consume_single_or_multi_attribute_values(
        self,
        attribute_name: DeviceAttributeName
    ) -> List[Union[DeviceId, DeviceInterfaceType, str]]:
        if self._value.startswith('{'):
            return self._consume_multi_attributes_values(attribute_name)
        else:
            return [self._consume_single_attribute_value(attribute_name)]

    def _consume_multi_attributes_values(
        self,
        attribute_name: DeviceAttributeName
    ) -> List[Union[DeviceId, DeviceInterfaceType, str]]:
        self._yank_expected('{')
        self._consume_optional_whitespaces()

        values = []

        while self._value and not self._value.startswith('}'):
            value = self._consume_single_attribute_value(attribute_name)
            values.append(value)
            self._consume_optional_whitespaces()

        if not values:
            self._raise_error('multi-valued attribute has no values')

        self._yank_expected('}')
        return values

    def _consume_single_attribute_value(
        self,
        attribute_name: DeviceAttributeName
    ) -> Union[DeviceId, DeviceInterfaceType, str]:

        if attribute_name is DeviceAttributeName.ID:
            return self._consume_usb_device_id()
        elif attribute_name is DeviceAttributeName.WITH_INTERFACE:
            return self._consume_interface_type()
        else:
            return self._consume_string_attribute_value()

    def _consume_usb_device_id(self) -> DeviceId:
        vendor_id = self._consume_hex_number_or_asterisk(4)
        self._yank_expected(':')

        if vendor_id is None:
            self._yank_expected('*')
            product_id = None
        else:
            product_id = self._consume_hex_number_or_asterisk(4)

        return DeviceId(vendor_id, product_id)

    def _consume_interface_type(self) -> DeviceInterfaceType:
        iface_class = self._consume_hex_number(2)
        self._yank_expected(':')
        iface_subclass = self._consume_hex_number_or_asterisk(2)
        self._yank_expected(':')

        if iface_subclass is None:
            self._yank_expected('*')
            iface_protocol = None
        else:
            iface_protocol = self._consume_hex_number_or_asterisk(2)

        return DeviceInterfaceType(iface_class, iface_subclass, iface_protocol)

    def _consume_string_attribute_value(self) -> str:
        self._yank_expected('"')
        quotes_index = self._value.find('"')
        if quotes_index == -1:
            self._raise_error('cannot find end of string')
        content = self._yank(quotes_index)
        self._yank_expected('"')
        return content

    def _consume_hex_number_or_asterisk(self, length: int) -> Optional[int]:
        if self._value.startswith('*'):
            self._yank()
            return None

        return self._consume_hex_number(length)

    def _consume_hex_number(self, length: int) -> int:
        hex_number = self._yank(length)
        try:
            return int(hex_number, 16)
        except ValueError as error:
            return self._raise_error(str(error))

    def _consume_reserved_word(self) -> str:
        length = 0
        for char in self._value:
            if char in self._RESERVED_WORD_CHARS:
                length += 1
            else:
                break

        if length == 0:
            self._raise_error('reserved word expected')

        return self._yank(length)

    def _can_consume_exact_word(self, w: str) -> bool:
        value = self._value
        word_length = len(w)
        return value == w \
            or (len(value) > word_length
                and value.startswith(w)
                and value[word_length] in whitespace)

    def _consume_mandatory_whitespaces(self) -> None:
        if not self._value or self._value[0] not in whitespace:
            self._raise_error('whitespace expected')
        self._consume_optional_whitespaces()

    def _consume_optional_whitespaces(self) -> None:
        self._value = self._value.lstrip(whitespace)

    def _yank_expected(self, value: str):
        if not self._value.startswith(value):
            self._raise_error(f'"{value}" expected')
        self._yank(len(value))

    def _yank(self, length: int = 1) -> str:
        if length > len(self._value):
            self._raise_error('unexpected end of string')
        yanked_part = self._value[:length]
        self._value = self._value[length:]
        return yanked_part

    def _raise_error(self, message: str):
        pos = len(self._input_value) - len(self._value)
        raise RuleParsingError(f'parsing error at position {pos}: {message}')



# USBGuard Simple GUI Py/Qt
# Copyright (C) 2019  Marco Nicola
#
# This file is part of "USBGuard Simple GUI Py/Qt".
#
# "USBGuard Simple GUI Py/Qt" is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# "USBGuard Simple GUI Py/Qt" is distributed in the hope that it will be
# useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with "USBGuard Simple GUI Py/Qt".  If not, see
# <https://www.gnu.org/licenses/>.

APP_NAME = 'USBGuard Simple GUI'



# USBGuard Simple GUI Py/Qt
# Copyright (C) 2019  Marco Nicola
#
# This file is part of "USBGuard Simple GUI Py/Qt".
#
# "USBGuard Simple GUI Py/Qt" is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# "USBGuard Simple GUI Py/Qt" is distributed in the hope that it will be
# useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with "USBGuard Simple GUI Py/Qt".  If not, see
# <https://www.gnu.org/licenses/>.

import signal
import sys

from PySide2.QtCore import QSize, QItemSelection
from PySide2.QtGui import Qt
from PySide2.QtWidgets import (QApplication,
                               QDialog,
                               QTableView,
                               QHeaderView,
                               QHBoxLayout,
                               QAbstractItemView,
                               QStyle,
                               QVBoxLayout,
                               QPushButton,
                               QWidget)
from . import APP_NAME
from .device import Device, DeviceModel
from .rules import RuleTarget
from .usbguard_dbus_interface import (CallbackEventType,
                                      EventPresenceChangeType,
                                      UsbguardDbusInterface)


class MainWindow(QDialog):
    def __init__(
        self,
        app: QApplication,
        usbguard_dbus: UsbguardDbusInterface
    ) -> None:
        super().__init__()

        self._app = app
        self._usbguard_dbus = usbguard_dbus

        self._register_dbus_callbacks()
        self._device_model = DeviceModel(self._usbguard_dbus.list_devices())
        self._device_table = self._create_device_table()
        self._controls_section = self._create_controls_section()
        self._init_window_content_and_aspect()

    def _register_dbus_callbacks(self) -> None:
        self._usbguard_dbus.register_callback(
            CallbackEventType.DEVICE_PRESENCE_CHANGED,
            self._on_device_presence_changed)

        self._usbguard_dbus.register_callback(
            CallbackEventType.DEVICE_POLICY_CHANGED,
            self._on_device_policy_changed)

    def _create_device_table(self) -> QTableView:
        device_table = QTableView()
        device_table.setModel(self._device_model)
        device_table.setSelectionBehavior(QAbstractItemView.SelectRows)
        device_table.setSelectionMode(QAbstractItemView.SingleSelection)

        device_table.selectionModel().selectionChanged.connect(
            self._on_device_table_selection_changed)

        horizontal_header = device_table.horizontalHeader()
        vertical_header = device_table.verticalHeader()
        horizontal_header.setSectionResizeMode(QHeaderView.ResizeToContents)
        vertical_header.setSectionResizeMode(QHeaderView.ResizeToContents)
        horizontal_header.setStretchLastSection(True)

        return device_table

    def _create_controls_section(self) -> QWidget:
        btn_allow = QPushButton('Allow')
        btn_allow.clicked.connect(self._on_allow_click)

        btn_block = QPushButton('Block')
        btn_block.clicked.connect(self._on_block_click)

        btn_reject = QPushButton('Reject')
        btn_reject.clicked.connect(self._on_reject_click)

        layout = QHBoxLayout()
        layout.addWidget(btn_allow)
        layout.addWidget(btn_block)
        layout.addWidget(btn_reject)
        layout.setContentsMargins(0, 0, 0, 0)

        controls_section = QWidget()
        controls_section.setLayout(layout)
        controls_section.hide()  # will show only on selection

        return controls_section

    def _init_window_content_and_aspect(self) -> None:
        window_layout = QVBoxLayout()
        window_layout.addWidget(self._device_table)
        window_layout.addWidget(self._controls_section)
        self.setLayout(window_layout)

        screen_geom = self._app.desktop().availableGeometry(self)
        window_size = QSize(screen_geom.width(), screen_geom.height()) * 0.8
        self.setGeometry(QStyle.alignedRect(
            Qt.LeftToRight, Qt.AlignCenter, window_size, screen_geom))

        self.setWindowTitle(APP_NAME)

    def _on_device_table_selection_changed(
        self,
        selected: QItemSelection,
        _deselected: QItemSelection
    ) -> None:
        self._controls_section.setVisible(not selected.isEmpty())

    def _on_device_presence_changed(
        self,
        device: Device,
        event: EventPresenceChangeType,
        _target: int
    ) -> None:
        if event is EventPresenceChangeType.REMOVE:
            self._device_model.remove_device(device)
        else:
            self._device_model.update_or_add_device(device)

    def _on_device_policy_changed(
        self,
        device: Device,
        _target_old: RuleTarget,
        _target_new: RuleTarget,
        _rule_id: int
    ) -> None:
        self._device_model.update_or_add_device(device)

    def _on_allow_click(self):
        device = self._get_selected_device()
        self._usbguard_dbus.apply_device_policy(
            device.device_id, RuleTarget.ALLOW, False)

    def _on_block_click(self):
        device = self._get_selected_device()
        self._usbguard_dbus.apply_device_policy(
            device.device_id, RuleTarget.BLOCK, False)

    def _on_reject_click(self):
        device = self._get_selected_device()
        self._usbguard_dbus.apply_device_policy(
            device.device_id, RuleTarget.REJECT, False)

    def _get_selected_device(self) -> Device:
        selection = self._device_table.selectionModel().selection()
        row = selection.first().topLeft().row()
        device = self._device_model.devices[row]
        return device


def main() -> None:
    signal.signal(signal.SIGINT, signal.SIG_DFL)

    app = QApplication(sys.argv)

    usbguard_dbus = UsbguardDbusInterface()

    main_window = MainWindow(app, usbguard_dbus)
    main_window.show()

    sys.exit(app.exec_())


if __name__ == '__main__':
    main()



# USBGuard Simple GUI Py/Qt
# Copyright (C) 2019  Marco Nicola
#
# This file is part of "USBGuard Simple GUI Py/Qt".
#
# "USBGuard Simple GUI Py/Qt" is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# "USBGuard Simple GUI Py/Qt" is distributed in the hope that it will be
# useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with "USBGuard Simple GUI Py/Qt".  If not, see
# <https://www.gnu.org/licenses/>.

from dataclasses import dataclass
from typing import Any, List, Optional
from PySide2.QtCore import QAbstractTableModel, QModelIndex, Qt
from .rules import DeviceAttribute, DeviceAttributeName, Rule


@dataclass
class Device:
    device_id: int
    rule: Rule

    @property
    def human_readable_name(self) -> str:
        name = self.rule.name
        if name and len(name.values) == 1:
            return name.values[0]
        return f'#{self.device_id}'


class DeviceModel(QAbstractTableModel):
    _ATTRIBUTES = [
        DeviceAttributeName.ID,
        DeviceAttributeName.NAME,
        DeviceAttributeName.WITH_CONNECT_TYPE,
        DeviceAttributeName.VIA_PORT,
        DeviceAttributeName.HASH,
        DeviceAttributeName.PARENT_HASH,
        DeviceAttributeName.SERIAL,
        DeviceAttributeName.WITH_INTERFACE,
    ]

    _HEADER = [
        'Target',
        'ID',
        'Name',
        'With Connect Type',
        'Via Port',
        'Hash',
        'Parent Hash',
        'Serial',
        'With Interface',
    ]

    def __init__(self, devices: List[Device]) -> None:
        super().__init__()
        self.devices: List[Device] = devices

    def update_or_add_device(self, device: Device) -> None:
        row = self._find_row_by_device_id(device.device_id)
        if row is None:
            self._add_new_device(device)
        else:
            self._update_device_at_row(device, row)

    def remove_device(self, device: Device) -> None:
        row = self._find_row_by_device_id(device.device_id)

        if row is not None:
            del self.devices[row]
            self.headerDataChanged.emit(Qt.Vertical, row, row)
            self.dataChanged.emit(row, row)

    def rowCount(self, parent: QModelIndex = QModelIndex()) -> int:
        return len(self.devices)

    def columnCount(self, parent: QModelIndex = QModelIndex()) -> int:
        return len(self._HEADER)

    def headerData(
        self,
        section: int,
        orientation: Qt.Orientation,
        role: int = Qt.DisplayRole
    ) -> Any:
        if role != Qt.DisplayRole:
            return None

        if orientation == Qt.Horizontal:
            return self._HEADER[section]
        elif section < len(self.devices):
            device = self.devices[section]
            return device.device_id

        return None

    def data(self, index: QModelIndex, role: int = Qt.DisplayRole) -> Any:
        if role != Qt.DisplayRole:
            return None

        row = index.row()
        column = index.column()
        device = self.devices[row]

        if column == 0:
            return device.rule.target.value
        else:
            attribute = self._ATTRIBUTES[column - 1]
            value = device.rule.attributes.get(attribute)
            return self._attribute_repr(value)

    def _add_new_device(self, device: Device) -> None:
        row = len(self.devices)
        self.devices.append(device)
        self.headerDataChanged.emit(Qt.Vertical, row, row)
        self.dataChanged.emit(row, row)

    def _update_device_at_row(self, device: Device, row: int) -> None:
        self.devices[row] = device
        self.dataChanged.emit(row, row)

    def _find_row_by_device_id(self, device_id: int) -> Optional[int]:
        for row, device in enumerate(self.devices):
            if device.device_id == device_id:
                return row
        return None

    @staticmethod
    def _attribute_repr(value: Optional[DeviceAttribute]) -> str:
        if value is None:
            return ''

        operator = value.operator
        values = value.values

        if len(values) == 1 and operator is None:
            return str(values[0])
        elif operator is None:
            return '\n'.join(map(str, values))
        else:
            values_repr = '\n'.join(f'    {str(v)}' for v in values)
            return f'{operator.value}:\n{values_repr}'



from setuptools import setup

setup(
    name='usbguard-simple-gui-py-qt',
    version='0.0.1',

    author='Marco Nicola',
    author_email='marconicola@disroot.org',

    description='A simple unambitious system tray applet and GUI for '
                'interacting with USBGuard.',
    long_description=open('README.md').read(),
    long_description_content_type="text/markdown",

    keywords='usbguard',
    url='https://github.com/marco-nicola/usbguard-simple-gui-py-qt',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: X11 Applications :: Qt',
        'Intended Audience :: End Users/Desktop',
        'License :: OSI Approved :: GNU General Public License v3 or later '
            '(GPLv3+)',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.7'
    ],

    packages=['usbguard_simple_gui_py_qt'],
    entry_points={
        'gui_scripts': [
            ('usbguard-simple-gui-py-qt = ' 
                'usbguard_simple_gui_py_qt.system_tray_app:main'),
            ('usbguard-simple-gui-py-qt-window-only = ' 
                'usbguard_simple_gui_py_qt.main_window:main'),
        ]
    },
    package_data={
        '': ['*.svg']
    },
    python_requires='>=3.7.0,<3.8',
    install_requires=[
        'dbus-python>=1.2.12,<2',
        'PySide2>=5.13.2,<6',
    ]
)
